import Foundation
import simd

struct Transforms {
	static func rotation(radians: Float, axis: SIMD3<Float>) -> matrix_float4x4 {
		let unitAxis = normalize(axis)
		let ct = cosf(radians)
		let st = sinf(radians)
		let ci = 1 - ct
		let x = unitAxis.x, y = unitAxis.y, z = unitAxis.z
		return matrix_float4x4.init(columns:(vector_float4(    ct + x * x * ci, y * x * ci + z * st, z * x * ci - y * st, 0),
		                                     vector_float4(x * y * ci - z * st,     ct + y * y * ci, z * y * ci + x * st, 0),
		                                     vector_float4(x * z * ci + y * st, y * z * ci - x * st,     ct + z * z * ci, 0),
		                                     vector_float4(                  0,                   0,                   0, 1)))
	}
	
	static func translation(_ translationX: Float, _ translationY: Float, _ translationZ: Float) -> matrix_float4x4
	{
		return matrix_float4x4.init(columns:(vector_float4(1, 0, 0, 0),
		                                     vector_float4(0, 1, 0, 0),
		                                     vector_float4(0, 0, 1, 0),
		                                     vector_float4(translationX, translationY, translationZ, 1)))
	}
	
	static func translate(_ translation: simd_float3) -> matrix_float4x4
	{
		return matrix_float4x4.init(columns:(vector_float4(1, 0, 0, 0),
		                                     vector_float4(0, 1, 0, 0),
		                                     vector_float4(0, 0, 1, 0),
		                                     vector_float4(translation.x, translation.y, translation.z, 1)))
	}

	static func perspective_fov(fovy: Float, aspectRatio: Float, nearZ: Float, farZ: Float) -> matrix_float4x4 {
		let radians = radians_from_degrees(0.5*fovy)
		let ys = 1 / tanf(radians)
		let xs = ys / aspectRatio
		let zs = farZ / (farZ - nearZ)
		return matrix_float4x4.init(columns:(vector_float4(xs,  0, 0,   0),
		                                     vector_float4( 0, ys, 0,   0),
		                                     vector_float4( 0,  0, zs,  1),
		                                     vector_float4( 0,  0, zs * -nearZ, 0)))
	}
	
	static func radians_from_degrees(_ degrees: Float) -> Float {
		return (degrees / 180) * .pi
	}
	
	static func lookAt(eye:simd_float3, center:simd_float3, up:simd_float3) -> matrix_float4x4 {
		let E:simd_float3 = -eye
		let N:simd_float3 = simd_normalize(center + E)
		let U:simd_float3 = simd_normalize(simd_cross(up,N))
		let V:simd_float3 = simd_cross(N,U)
		
		let P:simd_float4 = simd_float4(U.x,V.x,N.x,0)
		let Q:simd_float4 = simd_float4(U.y,V.y,N.y,0)
		let R:simd_float4 = simd_float4(U.z,V.z,N.z,0)
		let S:simd_float4 = simd_float4(simd_dot(U,E),simd_dot(V,E),simd_dot(N,E),1)
		
		return matrix_float4x4(P,Q,R,S)
	}
	
	static func scale(scale:simd_float3) -> matrix_float4x4 {
		let P:simd_float4 = simd_float4(scale.x,0,0,0)
		let Q:simd_float4 = simd_float4(0,scale.y,0,0)
		let R:simd_float4 = simd_float4(0,0,scale.z,0)
		let S:simd_float4 = simd_float4(0,0,0,1)
		return matrix_float4x4(P,Q,R,S)
	}
	
	static func rotateX(degree:Float) ->matrix_float4x4 {
		var P:simd_float4 = simd_float4(1,0,0,0)
		var Q:simd_float4 = simd_float4(0,1,0,0)
		var R:simd_float4 = simd_float4(0,0,1,0)
		var S:simd_float4 = simd_float4(0,0,0,1)
		let angle:Float = radians_from_degrees(degree)
		let s:Float = sin(angle)
		let c:Float = cos(angle)
		let a10:Float = 0
		let a11:Float = 1
		let a12:Float = 0
		let a13:Float = 0
		let a20:Float = 0
		let a21:Float = 0
		let a22:Float = 1
		let a23:Float = 0
		
		P.y = a10 * c + a20 * s
		Q.y = a11 * c + a21 * s
		R.y = a12 * c + a22 * s
		S.y = a13 * c + a23 * s
		P.z = a10 * -s + a20 * c
		Q.z = a11 * -s + a21 * c
		R.z = a12 * -s + a22 * c
		S.z = a13 * -s + a23 * c
		
		return matrix_float4x4(P,Q,R,S)
	}
	
	static func rotateY(degree:Float) ->matrix_float4x4 {
		var P:simd_float4 = simd_float4(1,0,0,0)
		var Q:simd_float4 = simd_float4(0,1,0,0)
		var R:simd_float4 = simd_float4(0,0,1,0)
		var S:simd_float4 = simd_float4(0,0,0,1)
		let angle:Float = radians_from_degrees(degree)
		let s:Float = sin(angle)
		let c:Float = cos(angle)
		let a00:Float = 1
		let a01:Float = 0
		let a02:Float = 0
		let a03:Float = 0
		let a20:Float = 0
		let a21:Float = 0
		let a22:Float = 1
		let a23:Float = 0
		
		P.x = a00 * c + a20 * -s
		Q.x = a01 * c + a21 * -s
		R.x = a02 * c + a22 * -s
		S.x = a03 * c + a23 * -s
		P.z = a00 * s + a20 * c
		Q.z = a01 * s + a21 * c
		R.z = a02 * s + a22 * c
		S.z = a03 * s + a23 * c
		
		return matrix_float4x4(P,Q,R,S)
	}
	
	static func rotateZ(degree:Float) ->matrix_float4x4 {
		var P:simd_float4 = simd_float4(1,0,0,0)
		var Q:simd_float4 = simd_float4(0,1,0,0)
		var R:simd_float4 = simd_float4(0,0,1,0)
		var S:simd_float4 = simd_float4(0,0,0,1)
		let angle:Float = radians_from_degrees(degree)
		let s:Float = sin(angle)
		let c:Float = cos(angle)
		let a00:Float = 1
		let a01:Float = 0
		let a02:Float = 0
		let a03:Float = 0
		let a10:Float = 0
		let a11:Float = 1
		let a12:Float = 0
		let a13:Float = 0
		
		P.x = a00 * c + a10 * s
		Q.x = a01 * c + a11 * s
		R.x = a02 * c + a12 * s
		S.x = a03 * c + a13 * s
		P.y = a00 * -s + a10 * c
		Q.y = a01 * -s + a11 * c
		R.y = a02 * -s + a12 * c
		S.y = a03 * -s + a13 * c
		
		return matrix_float4x4(P,Q,R,S)
	}

	static func toFloat3x3(f44:simd_float4x4) -> simd_float3x3 {
		let column0 = convertToFloat3 ( float4: f44.columns.0 )
		let column1 = convertToFloat3 ( float4: f44.columns.1 )
		let column2 = convertToFloat3 ( float4: f44.columns.2 )
		
		return simd_float3x3.init(column0, column1, column2)
	}
	
	static func convertToFloat3(float4: simd_float4) -> simd_float3 {
		return simd_float3.init(float4.x, float4.y, float4.z)
	}
}
